/*
 * oscp-slave-mdx.c
 * Copyright (C) 2015 KotCzarny <tjosko@yahoo.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "oscp-slave.h"

#include <sys/types.h> //socketpair
#include <sys/socket.h> //socketpair
#include <signal.h> //signal
#include <sys/time.h> //setpriority
#include <sys/resource.h> //setpriority


#define mdx_dbuf_size 1024*2
char mdx_dbuf[mdx_dbuf_size];

int mdxstate=0; float mdxcur=0, mdxdur=0; int mdxi=0; int mdxsong=0;
char mdxfn[PATH_MAX]; int mdxfs=0;
float mdxdurmax=245;

int mdx_send[2], mdx_recv[2], mdx_recv2[2]; // fds
int mdx_pid=-1;

/////################
void mdx_show_supp() {
 printf("  [slave-mdx]: external: mdxplay");
 printf(", formats: mdx, pdx\n");
}

/////###########
//void mdx_flush() { lf("mdx_flush"); adev_flush(); }
/////###########
void mdx_init() { lf("mdx_init"); mdxstate=0; dlog(9, "mdx_init"); if(mdxstate>0) return;

// if(mdxc) dlog(9, "mdx_init: mdxc not null?");

 if(adev_init()<=0) { mdxstate=0; return; }

 mdxstate=1;
}
/////##########
void mdx_stop() { lf("mdx_stop");
 dlog(9, "mdx_stop");
 if(mdxstate<=1) return;

 if(mdxstate>=3) {
  if(mdx_send[1]>0) close(mdx_send[1]); if(mdx_recv[1]>0) close(mdx_recv[1]);
  if(mdx_recv2[1]>0) close(mdx_recv2[1]);
  if(mdx_pid>0) {
   if(kill(mdx_pid,0)==0) { dlog(9,"mdx_stop: Killing slave."); kill(mdx_pid,SIGTERM); usleep(100000); }
   if(kill(mdx_pid,0)==0) kill(mdx_pid,SIGKILL);
   mdx_pid=-1;
  }
 }

 mdxstate=1;
}
/////#########
void mdx_end() { lf("mdx_end");
 dlog(9, "mdx_end"); if(mdxstate>1) mdx_stop(); adev_end();

 mdxstate=0; 
}
/////##################
void mdx_die(char *msg) { lf2("mdx_die",msg); dlog(9, "mdx_die"); mdx_stop(); cs(msg); }



/////##########################
void mdx_update_info(char *fn) { lf("mdx_update_info");
 char lmsg[1024];
 int fs=get_filesize(fn);
 snprintf(lmsg,1023,"Title: %s\n", get_basename(fn)); cs(lmsg);
 snprintf(lmsg,1023,"Comment: %d, MDX\n", fs); cs(lmsg);

/*
 sidplay2_st_getinfo(mdx_st, &mdx_sti);
 if(!mdx_sti.formatString) { dlog(9,"empty info ?"); return; }

 mdxdur=mdxdurmax; sprintf(lmsg, "ANS_length=%.1f\n", mdxdur); cs(lmsg);

 snprintf(lmsg,1022,"Title: %s (%s)\n", mdx_sti.infoString[0], get_basename(fn)); cs(lmsg);
 snprintf(lmsg,1022,"Artist: %s\n", mdx_sti.infoString[1]); cs(lmsg);

// mdx_md5();
 char md5[36]; sidplay2_stm_createMD5 (mdx_st, md5);
// snprintf(lmsg,1023,"md5: %s, hash: %08x", md5, md5toint(md5)); dlog(9,lmsg);
 md5[0]=97+0;
 int hash=md5toint(md5);
 int dur=mdx_getdur(hash);
snprintf(lmsg,1023,"md5: %s, hash: %08x, dur: %d", md5, md5toint(md5), dur); dlog(9,lmsg);
*/
}

////########################
int mdx_exec(char *mod_path) {
 if(socketpair(AF_UNIX, SOCK_STREAM, 0, mdx_send)<0) return -1;
 if(socketpair(AF_UNIX, SOCK_STREAM, 0, mdx_recv)<0) return -1;
 if(socketpair(AF_UNIX, SOCK_STREAM, 0, mdx_recv2)<0) return -1;

 signal(SIGPIPE, SIG_IGN); signal(SIGCHLD, SIG_IGN); // Ignore signal for death of the child process

 mdx_pid = fork(); if(mdx_pid == 0) // I am a child!
 { char lmsg[128]; sprintf(lmsg, "mdx_exec: i am child (pid: %d)", mdx_pid); dlog(9,lmsg);

  // pipe in/output through socket to parent
//  dup2(STDOUT_FILENO, STDERR_FILENO);  //redirects stderr to stdout
  dup2(mdx_send[0], STDIN_FILENO);
  dup2(mdx_recv[0], STDOUT_FILENO);
  dup2(mdx_recv2[0], STDERR_FILENO);
  close(mdx_send[1]); close(mdx_recv[1]); close(mdx_recv2[1]);
  close(mdx_recv2[0]);

  execlp("mdxplay", "mdxplay", "-s", "--output-as-raw", "-l", "0", mod_path, NULL);
  dlog(0,"mdx_exec: execlp failed"); exit(-1);
 }

 // I am a parent!, use mdx_xxx[1] to communicate
 close(mdx_send[0]); close(mdx_recv[0]); close(mdx_recv2[0]);

 int ret=7; if(mdx_pid==-1) // child is dead
 { if(mdx_send[1]>0) close(mdx_send[1]); if(mdx_recv[1]>0) close(mdx_recv[1]);
  if(mdx_recv2[1]>0) close(mdx_recv2[1]); mdx_send[1]=mdx_recv[1]=mdx_recv2[1]=-1; ret=-1; }
 else // child is alive
 { setpriority(PRIO_PROCESS,mdx_pid,-1); ret=0; }
 usleep(200000);

 return ret;
}

void mdx_play();
/////################################
void mdx_load(char *input_filename0) { lfd("mdx_load:state",mdxstate);
 static char lmsg[4096]; lmsg[4095]=0;
 snprintf(lmsg, 4095, "Playing %s\n", input_filename0); cs(lmsg);

 if(input_filename0[0]==0) return;
 if(input_filename0[0]==*"\"" && input_filename0[1]==*"\"" && input_filename0[2]==0)
 { cs("EOF code: stop requested\n"); if(mdxstate>2) mdx_stop(); return; }

 if(mdxstate==0) mdx_init();
 else if(mdxstate>2) mdx_stop();

 mdxstate=2; mdxcur=0; mdxdur=0;
 char *input_filename=input_filename0;
// int ret=0;


 mdxfs=get_filesize(input_filename);
 if(mdxfs<=0) { mdx_die("EOF code: invalid module size\n"); return; }

 if(mdx_exec(input_filename)!=0) { mdx_die("mdx_load: mdx_exec failed"); return; }

 sfmt=AV_SAMPLE_FMT_S16;
      if(sfmt==AV_SAMPLE_FMT_U8 || sfmt==AV_SAMPLE_FMT_U8P){ sformat.bits=8; }
 else if(sfmt==AV_SAMPLE_FMT_S16 || sfmt==AV_SAMPLE_FMT_S16P){ sformat.bits=16; }
 else if(sfmt==AV_SAMPLE_FMT_S32 || sfmt==AV_SAMPLE_FMT_S32P){ sformat.bits=32; }
 else if(sfmt==AV_SAMPLE_FMT_FLT || sfmt==AV_SAMPLE_FMT_FLTP){ sformat.bits=32; }
 else if(sfmt==AV_SAMPLE_FMT_DBL || sfmt==AV_SAMPLE_FMT_DBLP){ sformat.bits=64; }
 else { mdx_die("EOF code: Unsupported format\n"); return; }

 mdxcur=-1;

 sformat.channels=2; //(mdxconf.playback==mdx_stereo?2:1);
 sformat.rate=48000; //mdxconf.frequency;
#ifdef LAO
 sformat.byte_format=AO_FMT_NATIVE; sformat.matrix=0;
#endif

 adev_setp(sformat.bits, &sformat.channels, &sformat.rate);
sprintf(lmsg, "AUDIO: %d Hz, %d ch, xxx, %d kbit\n", sformat.rate, sformat.channels, 0); cs(lmsg);

 mdx_update_info(input_filename);

 mdxstate=3; mdxi=1;
 mdx_play();
}
void oscp_mdx_check_input();
void oscp_mdx_parse(char *a);
/////################# check messages
void mdx_check_input() { lf("mdx_check_input"); static char lmsg[4096]; static int rc;
 static fd_set mdxfsa;  static struct timeval mdxtv,*tvp;
 if(mdxstate==4) { tvp=NULL; }
 else { mdxtv.tv_sec=0; mdxtv.tv_usec=1; tvp=&mdxtv; }

 FD_ZERO(&mdxfsa); if(parent_said>=0) FD_SET(parent_said, &mdxfsa);

 rc=select(FD_SETSIZE, &mdxfsa, NULL, NULL, tvp); if(rc==0) return;
 else if(rc<0) { sprintf(lmsg, "Info: mdxci rc<0, %d, %s\n", errno, strerror(errno)); cs(lmsg); }
 else if(rc>0) // check if input is available
 { fdreadline(parent_said, lmsg, 1020); oscp_mdx_parse(lmsg); }
}
/////###################
void mdx_pause_toggle() { lf("mdx_pause_toggle");
      if(mdxstate==3) { mdxstate=4; cs("===== PAUSE =====\n");
  do { oscp_mdx_check_input(); } while(mdxstate==4);
 }
 else if(mdxstate==4) { mdxstate=3; cs("===== RESUME =====\n"); }
 else cs("Info: pause called but mdxstate != (3 || 4)"); }

/////##############
void mdx_checksd() {}
/////##############
void mdx_checkmd() {}

int mdx_fd[2][3]={ {-1,-1,-1}, {-1,-1,-1} };



////#############################
int mdx_slurp(char *buf, int len) { lf("mdx_slurp");
 int i, ret=0, fd=mdx_recv[1]; if(fd<0) { dlog(9, "mdx_slurp: fd<0"); return 0; }

 fd_set fsa, fsr; struct timeval tv; tv.tv_sec=0; tv.tv_usec=1000*1000;
 FD_ZERO(&fsr); FD_ZERO(&fsa); FD_SET(fd, &fsa);
 ret=select(FD_SETSIZE, &fsa, NULL, NULL, &tv);

 if(ret<0) { dlog(9, "mdx_slurp: ret<0"); }
 else if(ret==0) { dlog(9, "mdx_slurp: ret==0"); }
 else for(i=0; i<FD_SETSIZE; ++i) if(FD_ISSET(i, &fsa)) {
  if(i==fd) { usleep(10000); ret=read(fd, buf, len); break; }
 }

 return ret;
}
/////###########
void mdx_play() { lf("mdx_play"); dlog(9, "mdx_play");
 if(mdxstate!=3) { cs("EOF code: can't play, not inited\n"); return; }
 float ot1=-7.0, ts;

 static char lmsg[128]; void *buf[1]; int gb;

 while(mdxstate==3 && (gb=mdx_slurp(mdx_dbuf, mdx_dbuf_size))>0)
 {
  ts=((float)gb)/sformat.rate/sformat.channels/2.0;
  mdxcur+=ts; if((int)mdxcur!=(int)ot1)
  { ot1=mdxcur; sprintf(lmsg, "A: %.1f 0 0 0 0 0 0 0\n", mdxcur); cs(lmsg); }

  buf[0]=mdx_dbuf; adev_out(buf, gb/2/sformat.channels, sformat.channels, sfmt);

  mdxi++; mdx_check_input();
  if(mdxcur>=mdxdurmax) mdx_stop();
 }

 adev_close();
 mdx_die("EOF code: finished decoding\n");
 lf("mdx_play:finished");
}

/////########################
void oscp_mdx_load(char *fn) { lf("oscp_mdx_load");
 mdx_load(fn); //mdx_play();
 lf("oscp_mdx_load:finished");
}
/////#####################
void oscp_mdx_seek(int a) { lf("oscp_mdx_seek"); if(a<=0) return;
 float ts=((float)mdx_dbuf_size)/sformat.rate/sformat.channels/2.0; float s0=mdxcur; int db=0;
 while((db=mdx_slurp(mdx_dbuf, mdx_dbuf_size))>0 && mdxcur<s0+a) { mdxcur+=ts; }
}
/////########################
void oscp_mdx_parse(char *a) { lf("oscp_mdx_parse"); char lmsg[4096];
// sprintf(lmsg,"ofparse start"); dlog(9,lmsg);
 int l=strlen(a); if(l<=0) { sprintf(lmsg,"ofparse: l==%d ?", l); dlog(9,lmsg); return; }
 if(strcmp(a,"q")==0 || strcmp(a,"quit")==0) { mdx_end(); exit(0);
 } else if(strncmp(a,"loadfile ",9)==0) { oscp_mdx_load(a+9);
 } else if(strncmp(a,"load ",5)==0) { oscp_mdx_load(a+5);
 } else if(strncmp(a,"pause",5)==0) { mdx_pause_toggle();
 } else if(strncmp(a,"get_property length",19)==0) {
  sprintf(lmsg, "ANS_length=%.1f\n", mdxdur); cs(lmsg);
 } else if(strncmp(a,"get_property time_pos",19)==0 || strncmp(a,"get_time_pos",12)==0) {
//  sprintf(lmsg, "A: %.1f 0 0 0 0 0 0 0 0 0\n", mdxcur); cs(lmsg);
 } else if(strncmp(a,"seek ",5)==0) { int o1=0;
  if(a[5]==*"-") { sscanf(a+6, "%d", &o1); o1*=-1; }
  else if(a[5]==*"+") { sscanf(a+6, "%d", &o1); }
  else { sscanf(a+5, "%d", &o1); }
  oscp_mdx_seek(o1);
 } else if(strncmp(a,"vis ",4)==0) { sscanf(a+4, "%d", &vis);
 } else if(strncmp(a,"lfdump",5)==0) { lfdump();
 } else { sprintf(lmsg, "ofparse unhandled: %s", a); dlog(9, lmsg); }
 lf("oscp_mdx_parse:finished");
}
/////####################### check stdin
void oscp_mdx_check_input() { lf("oscp_mdx_check_input");
 char lmsg[4096]; int i, rc; fd_set fsa;
 struct timeval *tvp=NULL; //, tv; tv.tv_sec=0; tv.tv_usec=10000*99; // i=(int)time(NULL);
 usleep(10000); FD_ZERO(&fsa); if(parent_said>=0) FD_SET(parent_said, &fsa);
 rc=select(FD_SETSIZE, &fsa, NULL, NULL, tvp); // block till input

 if(rc<0) { if(0) { exit(EXIT_FAILURE); } else usleep(10000*99); }
 else if(rc==0) { }
 else for(i=0; i<FD_SETSIZE; ++i) if(FD_ISSET(i, &fsa)) {
  if(i==parent_said) { fdreadline(i, lmsg, 1020); oscp_mdx_parse(lmsg); }
 }
 lf("oscp_mdx_check_input:finished");
}
////################
int oscp_mdx_main() { lf("oscp_mdx_main");
 cs("oscp-slave mdxplay-20070206\n");
 lf("oscp_mdx_main:mdx_init");
 mdx_init();
 lf("oscp_mdx_main:while");
 while(1) { oscp_mdx_check_input(); }
 lf("oscp_mdx_main:mdx_end");
 mdx_end();
 lf("oscp_mdx_main:finished");
 cs("EOF code: exit\n");
// lfdump();
 return 0;
}
